/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

/*
 * The memory managers specific to matrices/vectors, which is based on
 * the one described in the K&R book (Chap. 8.7).  Also, following the
 * IEEE standard, we assume that a zero value (0.0) in a matrix/vector
 * is represented by all zero bits.
 *
 * [Note]
 * When running with OpenMP (i.e. _OPENMP is defined), a matrix/vector
 * that is created by a thread (with alloc_matrix/vector()) should be
 * released by the same thread (with release_matrix/vector()).
 */

#ifdef _OPENMP
static M_HEADER **matrix_pool0;
static int *max_mpool_size0;
static M_HEADER **mp_free0;
static M_HEADER **mp_last0;

static V_HEADER **vector_pool0;
static int *max_vpool_size0;
static V_HEADER **vp_free0;
static V_HEADER **vp_last0;
#else
static M_HEADER *matrix_pool0;
static int max_mpool_size0;
static M_HEADER *mp_free0;
static M_HEADER *mp_last0;

static V_HEADER *vector_pool0;
static int max_vpool_size0;
static V_HEADER *vp_free0;
static V_HEADER *vp_last0;
#endif

#ifdef _OPENMP
/* `th' should be declared as a local variable */
#  define matrix_pool    (matrix_pool0[th])
#  define max_mpool_size (max_mpool_size0[th])
#  define mp_free        (mp_free0[th])
#  define mp_last        (mp_last0[th])
#  define vector_pool    (vector_pool0[th])
#  define max_vpool_size (max_vpool_size0[th])
#  define vp_free        (vp_free0[th])
#  define vp_last        (vp_last0[th])
#else
#  define matrix_pool    matrix_pool0
#  define max_mpool_size max_mpool_size0
#  define mp_free        mp_free0
#  define mp_last        mp_last0
#  define vector_pool    vector_pool0
#  define max_vpool_size max_vpool_size0
#  define vp_free        vp_free0
#  define vp_last        vp_last0
#endif

#ifdef _OPENMP
void initialize_matrix_pool(void)
{
  int nth = omp_get_max_threads();
  int th;

  matrix_pool0 = (M_HEADER **)xmalloc(sizeof(M_HEADER *)*nth);
  max_mpool_size0 = (int *)xmalloc(sizeof(int)*nth);
  mp_free0 = (M_HEADER **)xmalloc(sizeof(M_HEADER *)*nth);
  mp_last0 = (M_HEADER **)xmalloc(sizeof(M_HEADER *)*nth);

  for (th=0; th<nth; th++) {
    matrix_pool0[th] = (M_HEADER *)xmalloc(sizeof(M_HEADER)*M_POOL_SIZE);
    memset(matrix_pool0[th],0,sizeof(M_HEADER)*M_POOL_SIZE);

    max_mpool_size0[th] = M_POOL_SIZE;
    mp_free0[th] = NULL;
    mp_last0[th] = matrix_pool0[th];
  }
}
#else
void initialize_matrix_pool(void)
{
  matrix_pool0 = (M_HEADER *)xmalloc(sizeof(M_HEADER)*M_POOL_SIZE);
  memset(matrix_pool0,0,sizeof(M_HEADER)*M_POOL_SIZE);

  max_mpool_size0 = M_POOL_SIZE;
  mp_free0 = NULL;
  mp_last0 = matrix_pool0;
}
#endif

#ifdef _OPENMP
void initialize_vector_pool(void)
{
  int nth = omp_get_max_threads();
  int th;

  vector_pool0 = (V_HEADER **)xmalloc(sizeof(V_HEADER *)*nth);
  max_vpool_size0 = (int *)xmalloc(sizeof(int)*nth);
  vp_free0 = (V_HEADER **)xmalloc(sizeof(V_HEADER *)*nth);
  vp_last0 = (V_HEADER **)xmalloc(sizeof(V_HEADER *)*nth);

  for (th=0; th<nth; th++) {
    vector_pool0[th] = (V_HEADER *)xmalloc(sizeof(V_HEADER)*V_POOL_SIZE);
    memset(vector_pool0[th],0,sizeof(V_HEADER)*V_POOL_SIZE);

    max_vpool_size0[th] = V_POOL_SIZE;
    vp_free0[th] = NULL;
    vp_last0[th] = vector_pool0[th];
  }
}
#else
void initialize_vector_pool(void)
{
  vector_pool0 = (V_HEADER *)xmalloc(sizeof(V_HEADER)*V_POOL_SIZE);
  memset(vector_pool0,0,sizeof(V_HEADER)*V_POOL_SIZE);

  max_vpool_size0 = V_POOL_SIZE;
  vp_free0 = NULL;
  vp_last0 = vector_pool0;
}
#endif

#ifdef _OPENMP
void release_matrix_pool(void)
{
  int nth = omp_get_max_threads();
  int th;
  int n0,nunits;
  M_HEADER *mp;
  FLOAT_T **p;

  for (th=0; th<nth; th++) {
    mp = matrix_pool0[th];
    while (mp < mp_last) {
      n0 = mp->info.dim0;

      if (n0 > 0) {
        p = (FLOAT_T **)(mp + 1);
        xfree(p[0]);
        nunits =
          (n0 * sizeof(FLOAT_T *) + sizeof(M_HEADER) - 1)/sizeof(M_HEADER) + 1;
      }
      else nunits = 1;

      mp += nunits;
    }
    xfree(matrix_pool0[th]);
  }

  xfree(matrix_pool0);
  xfree(max_mpool_size0);
  xfree(mp_free0);
  xfree(mp_last0);
}
#else
void release_matrix_pool(void)
{
  int n0,nunits;
  M_HEADER *mp;
  FLOAT_T **p;

  mp = matrix_pool0;
  while (mp < mp_last) {
    n0 = mp->info.dim0;

    if (n0 > 0) {
      p = (FLOAT_T **)(mp + 1);
      xfree(p[0]);
      nunits =
        (n0 * sizeof(FLOAT_T *) + sizeof(M_HEADER) - 1)/sizeof(M_HEADER) + 1;
    }
    else nunits = 1;

    mp += nunits;
  }

  xfree(matrix_pool0);
  max_mpool_size0 = 0;
  mp_free0 = NULL;
  mp_last0 = NULL;
}
#endif

#ifdef _OPENMP
void release_vector_pool(void)
{
  int nth = omp_get_max_threads();
  int th;

  for (th=0; th<nth; th++)
    xfree(vector_pool0[th]);

  xfree(vector_pool0);
  xfree(max_vpool_size0);
  xfree(vp_free0);
  xfree(vp_last0);
}
#else
void release_vector_pool(void)
{
  xfree(vector_pool0);
  max_vpool_size0 = 0;
  vp_free0 = NULL;
  vp_last0 = NULL;
}
#endif

MATRIX alloc_matrix(int n0, int n1)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  int i,nunits;
  M_HEADER *mp,*mp_prev;
  FLOAT_T **p,*q;

#ifdef CHECK_DIM
  if (n0 <= 0 || n1 <= 0) DIE_DIM;
#endif

  nunits =
    (n0 * sizeof(FLOAT_T *) + sizeof(M_HEADER) - 1)/sizeof(M_HEADER) + 1;

  if (mp_free == NULL) {
    matrix_pool->info.next = mp_free = matrix_pool;
    matrix_pool->info.dim0 = 0;
    matrix_pool->info.dim1 = 0;
    mp_last = matrix_pool + 1;
  }

  mp_prev = mp_free;

  mp = mp_prev->info.next;
  while (1) {
    if (mp->info.dim0 == n0 && mp->info.dim1 == n1) {
      mp_prev->info.next = mp->info.next;
      mp_free = mp_prev;
      return (MATRIX)(mp + 1);
    }

    if (mp == mp_free) {
      if ((mp_last + nunits - matrix_pool) >= max_mpool_size)
        die("Ran out of memory for matrices (allocated size: %d bytes)\n",
            max_mpool_size);

      mp_last->info.dim0 = n0;
      mp_last->info.dim1 = n1;

      p = (FLOAT_T **)(mp_last + 1);
      q = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*n0*n1);
      for (i=0; i<n0; i++)
        p[i] = q + i * n1;

      mp_last += nunits;

      return (MATRIX)p;
    }

    mp_prev = mp;
    mp = mp->info.next;
  }
}

VECTOR alloc_vector(int n)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  V_HEADER *vp,*vp_prev;
  int nunits;
  VECTOR p;

#ifdef CHECK_DIM
  if (n <= 0) DIE_DIM;
#endif

  nunits = (n * sizeof(FLOAT_T) + sizeof(V_HEADER) - 1)/sizeof(V_HEADER) + 1;

  if (vp_free == NULL) {
    vector_pool->info.next = vp_free = vector_pool;
    vector_pool->info.dim = 0;
    vp_last = vector_pool + 1;
  }

  vp_prev = vp_free;

  vp = vp_prev->info.next;
  while (1) {
    if (vp->info.dim == n) {
      vp_prev->info.next = vp->info.next;
      vp_free = vp_prev;

      return (VECTOR)(vp + 1);
    }

    if (vp == vp_free) {
      if ((vp_last + nunits - vector_pool) >= max_vpool_size)
        die("Ran out of memory for vectors (allocated size: %d bytes)\n",
            max_vpool_size);

      vp_last->info.dim = n;
      p = (VECTOR)(vp_last + 1);
      vp_last += nunits;

      return p;
    }

    vp_prev = vp;
    vp = vp->info.next;
  }
}

void release_matrix0(MATRIX m)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  M_HEADER *mp,*b;

  if (m == NULL) return;

  b = (M_HEADER *)m - 1;

#ifdef CHECK_DIM
  if (b < matrix_pool || b > mp_last)
    die("Internal error: Illegal address of a matrix to be freed\n");
#endif

  mp = mp_free;
  while (1) {
    if (b == mp) die("Internal error: double-freed a matrix (%p)\n",b);
    if ((b > mp && b < mp->info.next) ||
        (mp >= mp->info.next && (b > mp || b < mp->info.next))) break;
    mp = mp->info.next;
  }

  b->info.next = mp->info.next;
  mp->info.next = b;

  mp_free = b;
}

void release_vector0(VECTOR v)
{
#ifdef _OPENMP
  int th = omp_get_thread_num();
#endif
  V_HEADER *vp,*b;

  if (v == NULL) return;

  b = (V_HEADER *)v - 1;

#ifdef CHECK_DIM
  if (b < vector_pool || b > vp_last)
    die("Internal error: Illegal address of a vector to be freed\n");
#endif

  vp = vp_free;
  while (1) {
    if (b == vp) die("Internal error: double-freed a vector (%p)\n",b);
    if ((b > vp && b < vp->info.next) ||
        (vp >= vp->info.next && (b > vp || b < vp->info.next))) break;
    vp = vp->info.next;
  }

  b->info.next = vp->info.next;
  vp->info.next = b;

  vp_free = b;
}
