/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef USE_ICONV
#  include <iconv.h>
#endif

#include "xgetopt.h"

#define TOOL_NAME "NBCT"
#ifndef VERSION
#  define VERSION "unknown"
#endif
#define MAX_IN_FILENAME  1024
#define MAX_EXTENSION    10
#define MAX_ID           1024
#define MAX_DFLT_LEN     16
#define DEFAULT_BASE     "out"
#define DEFAULT_EXT      "out"
#define CHARCODE_LEN      100
#define MAX_LINE_LEN     65536
#define MAX_OUT_LEN      65536
#define INIT_NUM_FILES   30
#ifdef USE_ICONV
#  define CHARCODE_INTERNAL "UTF-8"
#endif

enum match_type {MATCH_INFIX, MATCH_EXACT, MATCH_PREFIX, MATCH_SUFFIX};
enum nbc_option_id {
  O_SEP_DEL_HEAD=1, O_SEP_DEL_BODY, O_SEP_EXT, O_SEP_TARGET,
  O_SEP_HELP, O_SEP_MATCH, O_SEP_OUT, O_SEP_CODE_IN, O_SEP_CODE_OUT,
  O_SEP_SUFFIX, O_SEP_VIEW, O_SEP_VERSION, O_SEP_HIDE_FILE
};

OPTION nbcsep_opts[] = {
  { O_SEP_DEL_HEAD,            "del-head",            'd',  0 },
  { O_SEP_DEL_BODY,            "del-body",            'D',  0 },
  { O_SEP_EXT,                 "ext",                 'e',  1 },
  { O_SEP_TARGET,              "target",              'f',  1 },
  { O_SEP_HIDE_FILE,           "hide-file",           'F',  0 },
  { O_SEP_HELP,                "help",                'h',  0 },
  { O_SEP_MATCH,               "match",               'm',  1 },
  { O_SEP_OUT,                 "output",              'o',  1 },
#ifdef USE_ICONV
  { O_SEP_CODE_IN,             "code-in",             'p',  1 },
  { O_SEP_CODE_OUT,            "code-out",            'q',  1 },
#endif
  { O_SEP_SUFFIX,              "suffix",              's',  1 },
  { O_SEP_VIEW,                "view",                'v',  1 },
  { O_SEP_VERSION,             "version",               0,  0 },
  { -1, NULL, 0,  0 }
};

char **fname_input;
char fname_base[MAX_IN_FILENAME];
char fname_suffix[MAX_IN_FILENAME];
char fname_ext[MAX_EXTENSION];
char id_str[MAX_ID];
int mode_ext;
int mode_suffix;
int mode_id;
int mode_list;
int mode_stdout;
int mode_conv_in;
int mode_conv_out;
enum match_type mode_match;
int mode_del_head;
int mode_del_body;
int mode_hide_file;
#ifdef USE_ICONV
iconv_t cd_in, cd_out;
char charcode_in[CHARCODE_LEN];
char charcode_out[CHARCODE_LEN];
#endif
char *buf_mid;
char *buf_cout;
int num_files;
int max_num_files;
FILE **fp_input = NULL;

void print_version(FILE *fp)
{
  char version[] =
    TOOL_NAME " version " VERSION " (built on " __DATE__ ")\n";

  fputs(version,fp);
}

void print_usage(FILE *fp)
{
  char usage[] =
    TOOL_NAME " version " VERSION " (built on " __DATE__ ")\n"
    "[Usage] nbcsep OPTIONS FILE [FILE ...]\n"
    "Flags with an argument:\n"
    " --target (-f) Common prefix of the output file(s)\n"
    " --suffix (-s) Common suffix of the output file(s)\n"
    " --ext    (-e) File extension of the output file(s)\n"
    " --view   (-v) ID of the data matrix to display\n"
    " --output (-o) ID of the data matrix to save\n"
    " --match  (-m) Matching criterion for --view and --output\n"
    "              (exact, infix, prefix and suffix) [Default: infix]\n"
#ifdef USE_ICONV
#ifdef CHARCODE
    " --code-in  (-p) Input character code  [Default: " CHARCODE "]\n"
    " --code-out (-q) Output character code [Default: " CHARCODE "]\n"
#else
    " --code-in  (-p) Input character code  [Default: " CHARCODE_INTERNAL "]\n"
    " --code-out (-q) Output character code [Default: " CHARCODE_INTERNAL "]\n"
#endif
#endif
    "Flags with no argument:\n"
    " --del-head  (-d) Show only data matrices\n"
    " --del-body  (-D) Show only header parts\n"
    " --hide-file (-F) Hide filenames\n"
    " --help      (-h) Print this message\n";

  fprintf(fp,"\n%s\n",usage);
}

void initialize_modes(void)
{
  strncpy(fname_base,DEFAULT_BASE,MAX_DFLT_LEN);
  mode_ext = 0;
  mode_suffix = 0;
  mode_id = 0;
  mode_list = 1;
  mode_stdout = 0;
  mode_match = MATCH_INFIX;
  mode_del_head = 0;
  mode_del_body = 0;
  mode_hide_file = 0;
#ifdef USE_ICONV
#ifdef CHARCODE
  strncpy(charcode_in,CHARCODE,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE,CHARCODE_LEN);
#else
  strncpy(charcode_in,CHARCODE_INTERNAL,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE_INTERNAL,CHARCODE_LEN);
#endif
#endif
  num_files = 0;
  max_num_files = INIT_NUM_FILES;
}

void initialize_buffer(void)
{
  buf_mid = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  if (buf_mid == NULL) {
    fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
    exit(1);
  }
  buf_cout = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  if (buf_cout == NULL) {
    fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
    exit(1);
  }
  *buf_mid = '\0';
  *buf_cout = '\0';
}

void release_buffer(void)
{
  free(buf_mid);
  buf_mid = NULL;
  free(buf_cout);
  buf_cout = NULL;
}

#ifdef USE_ICONV
void initialize_iconv(void)
{
  cd_in = iconv_open(CHARCODE_INTERNAL,charcode_in);
  if (cd_in == (iconv_t)-1) {
    fprintf(stderr,
            "iconv_open failed (in: %s)"
            " -- run 'iconv -l' for available character codes\n",
            charcode_in);
    exit(1);
  }

  cd_out = iconv_open(charcode_out,CHARCODE_INTERNAL);
  if (cd_out == (iconv_t)-1) {
    fprintf(stderr,
            "iconv_open failed (out: %s)"
            " -- run 'iconv -l' for available character codes\n",
            charcode_out);
    exit(1);
  }
}
#endif

void alloc_fname_input(void)
{
  int n;

  fname_input = (char **)malloc(sizeof(char *)*max_num_files);
  if (fname_input == NULL) {
    fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
    exit(1);
  }

  for (n = 0; n < max_num_files; n++) {
    fname_input[n] = (char *)malloc(sizeof(char)*MAX_IN_FILENAME);
    if (fname_input[n] == NULL) {
      fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
      exit(1);
    }
    memset(fname_input[n],0,MAX_IN_FILENAME);
  }
}

void expand_fname_input(void)
{
  int old_max_num_files,n;

  old_max_num_files = max_num_files;
  max_num_files *= 2;

  fname_input = (char **)realloc(fname_input,sizeof(char *)*max_num_files);
  if (fname_input == NULL) {
    fprintf(stderr,"Couldn't reallocate memory (%s,%d)\n",__FILE__,__LINE__);
    exit(1);
  }

  for (n = old_max_num_files; n < max_num_files; n++) {
    fname_input[n] = (char *)malloc(sizeof(char)*MAX_IN_FILENAME);
    if (fname_input[n] == NULL) {
      fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
      exit(1);
    }
    memset(fname_input[n],0,MAX_IN_FILENAME);
  }
}

void release_fname_input(void)
{
  int n;

  for (n = 0; n < max_num_files; n++) free(fname_input[n]);
  free(fname_input);
  fname_input = NULL;
}

void initialize_all(void)
{
  initialize_modes();
  initialize_buffer();
#ifdef USE_ICONV
  initialize_iconv();
#endif

  alloc_fname_input();
}

void release_fp_input(void)
{
  int n;

  for (n = 0; n < num_files; n++)
    if (fp_input[n] != stdin) fclose(fp_input[n]);

  free(fp_input);
  fp_input = NULL;
}

/* [NOTE] this function only returns the pointer for a file extension,
 * without making a copy
 */
char *get_one_file_extension(char *fname)
{
  int i,len,has_dot;
  char *ext = NULL;

  len = strlen(fname);
  has_dot = 0;

  for (i=0; i<len; i++) {
    if (fname[i] == '.') {
      has_dot = 1;
      if (i < (len - 1)) {
        i++;
        ext = fname + i;
      }
      else {
        ext = "";
      }
    }
  }

  return (has_dot && ext != NULL) ? ext : DEFAULT_EXT;
}

void get_file_extension(void)
{
  int n;
  char *ext = NULL;
  char *last_ext = NULL;
  int has_common;

  has_common = 1;
  for (n = 0; n < num_files; n++) {
    if (strcmp(fname_input[n],"-") == 0) continue;

    ext = get_one_file_extension(fname_input[n]);
    if (n == 0) {
      last_ext = ext;
    }
    else if (strcmp(last_ext,ext) != 0) {
      has_common = 0;
      break;
    }
  }

  if (has_common)
    strncpy(fname_ext,last_ext,MAX_EXTENSION);
  else
    strncpy(fname_ext,DEFAULT_EXT,MAX_EXTENSION);
}

void parse_args(int argc, char *argv[])
{
  int c;
  int i,j;
  int n = 0;

  init_xgetopt();
  while (1) {
    c = xgetopt(argc,argv,nbcsep_opts);
    if (c == -1) break;

    switch (c) {
    case -2:
      fprintf(stderr,
              "Unknown/illegal option specification: %s (Use -h for usage)\n",
              xoptarg);
      exit(1);
    case O_SEP_DEL_HEAD:
      mode_del_head = 1;
      break;
    case O_SEP_DEL_BODY:
      mode_del_body = 1;
      break;
    case O_SEP_EXT:
      mode_ext = 1;
      copy_optarg(fname_ext,(MAX_EXTENSION - 1));
      break;
    case O_SEP_TARGET:
      copy_optarg(fname_base,(MAX_IN_FILENAME - 1));
      mode_list = 0;
      break;
    case O_SEP_HIDE_FILE:
      mode_hide_file = 1;
      break;
    case O_SEP_VERSION:
      print_version(stdout);
      exit(0);
      break;
    case O_SEP_HELP:
      print_usage(stdout);
      exit(0);
      break;
    case O_SEP_MATCH:
      for (j=0; j<strlen(xoptarg_c); j++)
        xoptarg_c[j] = toupper(xoptarg_c[j]); /* working on temporary area */
      if (strcmp(xoptarg_c,"INFIX") == 0) {
        mode_match = MATCH_INFIX;
      }
      else if (strcmp(xoptarg_c,"EXACT") == 0) {
        mode_match = MATCH_EXACT;
      }
      else if (strcmp(xoptarg_c,"PREFIX") == 0) {
        mode_match = MATCH_PREFIX;
      }
      else if (strcmp(xoptarg_c,"SUFFIX") == 0) {
        mode_match = MATCH_SUFFIX;
      }
      else {
        fprintf(stderr,
                "Unknown matching criterion -- %s (Use for usage)\n",
                xoptarg);
        exit(1);
      }
      break;
    case O_SEP_OUT:
      mode_id = 1;
      mode_list = 0;
      copy_optarg(id_str,(MAX_ID - 1));
      break;
#ifdef USE_ICONV
    case O_SEP_CODE_IN:
      copy_optarg(charcode_in,(CHARCODE_LEN - 1));
      break;
    case O_SEP_CODE_OUT:
      copy_optarg(charcode_out,(CHARCODE_LEN - 1));
      break;
#endif
    case O_SEP_SUFFIX:
      mode_suffix = 1;
      copy_optarg(fname_suffix,(MAX_IN_FILENAME - 1));
      break;
    case O_SEP_VIEW:
      mode_stdout = 1;
      mode_id = 1;
      mode_list = 0;
      copy_optarg(id_str,(MAX_ID - 1));
      break;
    case 0:
      if (n == max_num_files) expand_fname_input();
      copy_optarg(fname_input[n],(MAX_IN_FILENAME - 1));
      n++;
      break;
    default:
      break;
    }
  }

  num_files = n;
  if (num_files == 0) {
    fprintf(stderr,"No input file specified (Use -h for usage)\n");
    exit(1);
  }

#ifdef USE_ICONV
  mode_conv_in = (strcmp(charcode_in,CHARCODE_INTERNAL) == 0) ? 0 : 1;
  mode_conv_out = (strcmp(charcode_out,CHARCODE_INTERNAL) == 0) ? 0 : 1;
#else
  mode_conv_in = 0;
  mode_conv_out = 0;
#endif

  /* Convert the specified ID into the upper case */
  for (i = 0; i < strlen(id_str); i++) {
    id_str[i] = toupper(id_str[i]);
  }
}

void prepare_input(void)
{
  int n;

  fp_input = (FILE **)malloc(sizeof(FILE *)*num_files);
  if (fp_input == NULL) {
    fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
    exit(1);
  }

  for (n = 0; n < num_files; n++) {
    if (strcmp(fname_input[n],"-") == 0) {
      fp_input[n] = stdin;
    }
    else {
      fp_input[n] = fopen(fname_input[n],"rb");
      if (fp_input[n] == NULL) {
        fprintf(stderr, "Couldn't open %s\n", fname_input[n]);
        exit(1);
      }
    }
  }

  if (!mode_ext) get_file_extension();
}

char *get_comment_id(char *buf)
{
  char *comment_id;
  char *b,*q;
  int l;

  b = buf;

  if (*b == '#') b++;
  else return NULL;
  while (*b == '#') b++;

  if (*b == ' ') b++;
  else return NULL;
  while (*b == ' ') b++;

  if (*b == 'I') b++;
  else return NULL;
  if (*b == 'D') b++;
  else return NULL;
  if (*b == ':') b++;
  else return NULL;

  comment_id = (char *)malloc(sizeof(char)*(MAX_ID + 1));
  if (comment_id == NULL) {
    fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
    exit(1);
  }
  *comment_id = '\0';

  q = comment_id;
  l = 0;
  while (*b != '\0' && *b != ' ' && *b != '\t' &&
         *b != '\r' && *b != '\n') {
    *q = *b;
    if (l > MAX_ID - 1) {
      fprintf(stderr,"Internal error: too long data matrix ID\n");
      exit(1);
    }
    q++;
    b++;
    l++;
  }
  *q = '\0';

  return comment_id;
}

char *build_out_filename(char *comment_id)
{
  char *fname_out;
  int name_len;
  int c;

  name_len = strlen(fname_base);
  if (mode_suffix) name_len += strlen(fname_suffix) + 1;
  name_len += strlen(comment_id) + 2;
  name_len += (mode_ext) ? strlen(fname_ext) : strlen(DEFAULT_EXT);

  fname_out = (char *)malloc(sizeof(char)*(name_len + 1));
  if (fname_out == NULL) {
    fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
    exit(1);
  }
  *fname_out = '\0';

  for (c = 0; c < name_len + 1; c++)
    fname_out[c] = '\0';

  strncpy(fname_out,fname_base,MAX_IN_FILENAME);
  if (mode_suffix) {
    strncat(fname_out,"_",1);
    strncat(fname_out,fname_suffix,MAX_IN_FILENAME);
  }
  strncat(fname_out,"_",1);
  strncat(fname_out,comment_id,MAX_ID);
  strncat(fname_out,".",1);
  strncat(fname_out,fname_ext,MAX_EXTENSION);

  return fname_out;
}

/* Simple version of the one defined in misc.c */
char *conv_out(char *buf_in)
{
#ifdef USE_ICONV
  int buf_len_mid0;
  size_t buf_len_mid, buf_len_out;
  char *b;
#ifdef ICONV_CONST
  const char *bm_tmp;
#else
  char *bm_tmp;
#endif
#endif
#ifdef USE_CRLF
  char *bi,*bm;
  int l;
#endif

#ifdef USE_CRLF
  /* Conversion of the line separaters: LF -> CR+LF */
  *buf_mid = '\0';
  bi = buf_in;
  bm = buf_mid;
  l = 0;
  while (*bi != '\0') {
    if (l >= MAX_OUT_LEN) {
      fprintf(stderr,"Too long line to output: %s",buf_in);
      exit(1);
    }
    if (*bi == '\n') {
      *bm = '\r';
      *(bm+1) = '\n';
      bm += 2;
      l += 2;
    }
    else {
      *bm = *bi;
      bm++;
      l++;
    }
    bi++;
  }
  *bm = '\0';
#else
  strncpy(buf_mid,buf_in,MAX_OUT_LEN);
#endif

#ifdef USE_ICONV
  if (!mode_conv_out) return buf_mid;

  *buf_cout = '\0';
  b = buf_cout;
  bm_tmp = buf_mid;
  buf_len_mid0 = strlen(buf_mid);
  buf_len_mid = (size_t)buf_len_mid0;
  buf_len_out = (size_t)MAX_OUT_LEN;
  if (iconv(cd_out,&bm_tmp,&buf_len_mid,&b,&buf_len_out) == (size_t)-1) {
    fprintf(stderr,"Couldn't convert: %s (errno=%d)\n",buf_mid,errno);
    exit(1);
  }
  buf_cout[MAX_OUT_LEN - buf_len_out] = '\0';
#else
  return buf_mid;
#endif

  return buf_cout;
}

void filter_one_input(char *fname, FILE *fp_in)
{
  int line_c;
  FILE *fp_out = NULL;
  char *fname_out;
  char *comment_id;
  char has_out;
  char printable;
  char *buf;
  char *buf_conv;
#ifdef USE_ICONV
  char *bb;
  size_t buf_len_in, buf_len_out;
#  ifdef ICONV_CONST
  const char *buf_tmp = NULL;
#  else
  char *buf_tmp = NULL;
#  endif
#endif
  char *s;
  char is_comment;

  buf = (char *)malloc(sizeof(char)*(MAX_LINE_LEN + 1));
  if (buf == NULL) {
    fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
    exit(1);
  }
  *buf = '\0';

#ifdef USE_ICONV
  buf_conv = (char *)malloc(sizeof(char)*(MAX_LINE_LEN + 1));
  if (buf_conv == NULL) {
    fprintf(stderr,"Couldn't allocate memory (%s,%d)\n",__FILE__,__LINE__);
    exit(1);
  }
  *buf_conv = '\0';
#endif

  line_c = 0;
  has_out = 0;
  printable = 0;
  while (fgets(buf,MAX_LINE_LEN,fp_in) != NULL) {
    line_c++;

#ifdef USE_ICONV
    if (mode_conv_in) {
      buf_tmp = buf;
      bb = buf_conv;
      buf_len_in  = strlen(buf)+1;
      buf_len_out = MAX_LINE_LEN+1;
      if (iconv(cd_in,&buf_tmp,&buf_len_in,&bb,&buf_len_out) == (size_t)-1) {
        fprintf(stderr,
                "[line %d] Couldn't convert to the internal charcode"
                " (errno=%d)\n",
                line_c,errno);
        exit(1);
      }
    }
    else buf_conv = buf;
#else
    buf_conv = buf;
#endif

    comment_id = get_comment_id(buf_conv);

    if (comment_id != NULL) {

      if (mode_list) {
        if (num_files > 1 && !mode_hide_file) {
          if (fp_in == stdin) {
            fprintf(stdout,"(stdin): %s\n",comment_id);
          }
          else {
            fprintf(stdout,"%s: %s\n",fname,comment_id);
          }
        }
        else {
          fprintf(stdout,"%s\n",comment_id);
        }

        free(comment_id);
        comment_id = NULL;
        continue;
      }

      if (!mode_stdout && has_out) fclose(fp_out);

      if (!mode_id ||
          (mode_match == MATCH_INFIX  && strstr(comment_id,id_str) != NULL) ||
          (mode_match == MATCH_EXACT  && strcmp(comment_id,id_str) == 0)    ||
          (mode_match == MATCH_PREFIX &&
           (s = strstr(comment_id,id_str)) != NULL && comment_id == s) ||
          (mode_match == MATCH_SUFFIX &&
           (s = strstr(comment_id,id_str)) != NULL &&
           strlen(id_str) == strlen(s))) {

        if (mode_stdout) {
          if (!mode_del_head) {
            if (num_files > 1 && !mode_hide_file) {
              if (fp_in == stdin) {
                fputs("(stdin): ", stdout);
              }
              else {
                fprintf(stdout, "%s: ", fname);
              }
            }
            fputs(conv_out(buf_conv),stdout);
          }
          has_out = 0;
        }
        else {
          fname_out = build_out_filename(comment_id);

          if (!mode_del_body) {
            fprintf(stderr,"Output: ");
            if (num_files > 1 && !mode_hide_file) {
              if (fp_in == stdin) {
                fprintf(stderr, "(stdin): ");
              }
              else {
                fprintf(stderr, "%s: ", fname);
              }
            }
            fprintf(stderr,"%s\n",fname_out);
          }

          fp_out = fopen(fname_out,"wb");
          if (fp_out == NULL) {
            fprintf(stderr,"Couldn't open %s\n", fname_out);
            exit(1);
          }

          has_out = 1;

          fputs(conv_out(buf_conv),fp_out);
          free(fname_out);
          fname_out = NULL;
        }

        printable = 1;
      }
      else {
        has_out = 0;
        printable = 0;
      }

      free(comment_id);
      comment_id = NULL;
    }
    else if (printable) {
      is_comment = (strlen(buf_conv) > 0 && buf_conv[0] == '#');
      if ((!mode_del_head && is_comment) || (!mode_del_body && !is_comment)) {
        if (has_out)
          fputs(conv_out(buf_conv),fp_out);
        else {
          if (num_files > 1 && !mode_hide_file) {
            if (fp_in == stdin) {
              fputs("(stdin): ", stdout);
            }
            else {
              fprintf(stdout, "%s: ", fname);
            }
          }
          fputs(conv_out(buf_conv),stdout);
        }
      }
    }
  }

  if (!mode_stdout && has_out) fclose(fp_out);

  free(buf);
  buf = NULL;
#ifdef USE_ICONV
  if (mode_conv_in) {
    free(buf_conv);
    buf_conv = NULL;
  }
#endif
}

void filter_input(void)
{
  int n;

  for (n = 0; n < num_files; n++) {
    filter_one_input(fname_input[n], fp_input[n]);
  }
}

void finalize_all(void)
{
  release_buffer();
#ifdef USE_ICONV
  iconv_close(cd_in);
  iconv_close(cd_out);
#endif

  release_fname_input();
  release_fp_input();
}

int main(int argc, char *argv[])
{
  initialize_all();
  parse_args(argc, argv);
  prepare_input();
  filter_input();
  finalize_all();
  return 0;
}
