/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OToHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

FLOAT_T lwishart_coeff(int n, FLOAT_T alpha)
{
  int i;
  FLOAT_T c;

  c = (alpha * n) * 0.5 * LOG(2.0);
  c += n * (n-1) * 0.25 * LOG(M_PI);

  for (i=1; i<=n; i++) /* [Note] i should be 1...n by definition */
    c += LGAMMA((alpha + 1.0 - i) * 0.5);

  return -c;
}

FLOAT_T
lwishart_det(MATRIX p_p, MATRIX h_s, MATRIX h_p,
             FLOAT_T p_p_det, FLOAT_T h_p_det, FLOAT_T alpha)
{
  FLOAT_T d;
  int n;
  MATRIX x;

#ifdef CHECK_DIM
  if ((n = dim0(p_p)) <= 0 || n != dim1(p_p) ||
      n != dim0(h_p) || n != dim1(h_p))
    DIE_DIM;
#else
  n = dim0(p_p);
#endif

  x = alloc_matrix(n,n);
  prod_matrix(h_s,p_p,x);

  d = lwishart_coeff(n,alpha);
  d -= 0.5 * alpha * LOG(h_p_det);
  d += 0.5 * (alpha - n - 1) * LOG(p_p_det);
  d -= 0.5 * trace(x);

  release_matrix(x);

  return d;
}

FLOAT_T lwishart(MATRIX p_p, MATRIX h_p, FLOAT_T alpha)
{
  FLOAT_T d,p_p_det,h_p_det;
  int n;
  MATRIX h_s;

#ifdef CHECK_DIM
  if ((n = dim0(p_p)) <= 0 || n != dim1(p_p) ||
      n != dim0(h_p) || n != dim1(h_p))
    DIE_DIM;
#else
  n = dim0(p_p);
#endif

  h_s = alloc_matrix(n,n);
  INVERT(h_p,h_s);

  p_p_det = DETERMINANT(p_p);
  h_p_det = DETERMINANT(h_p);

  d = lwishart_det(p_p,h_p,h_s,p_p_det,h_p_det,alpha);

  release_matrix(h_s);

  return d;
}

FLOAT_T
wishart_det(MATRIX p_p, MATRIX h_p, MATRIX h_s,
            FLOAT_T p_p_det, FLOAT_T h_p_det, FLOAT_T alpha)
{
  return EXP(lwishart_det(p_p,h_s,h_p,p_p_det,h_p_det,alpha));
}

FLOAT_T wishart(MATRIX p_p, MATRIX h_p, FLOAT_T alpha)
{
  return EXP(lwishart(p_p,h_p,alpha));
}

